/*
/--------------------------------------------------------------------
|
|      $Id: jpegdec.cpp,v 1.5 1999/10/19 21:25:16 Ulrich von Zadow Exp $
|      JPEG Decoder Class
|
|      JPEG file decoder. Uses the independent JPEG group's library
|      to do the actual conversion.
|
|      Copyright (c) 1996-1998 Ulrich von Zadow
|
\--------------------------------------------------------------------
*/

#include "stdpch.h"
#define JPEG_INTERNALS
#include "jpegdec.h"

#include "except.h"

extern "C"
{
#include "jmemsrc.h"
}

#ifdef _WINDOWS
#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif
#endif

// Check to see if the pixel format defined in jmorecfg.h matches the
// one defined in bitmap.h.
// If the program reports an error here, you're probably using a wrong
// version of jmorecfg.h. Several files in libjpeg and libtiff need to
// replaced for the libraries to be used with paintlib. See the
// documentation for details.
#if (RGBA_RED != RGB_RED) || (RGBA_GREEN != RGB_GREEN)   \
    || (RGBA_BLUE != RGB_BLUE) || (RGB_PIXELSIZE != PIXEL_SIZE)
#error Illegal JPEG pixel format
#endif

IMPLEMENT_DYNAMIC (CJPEGDecoder, CPicDecoder);

/////////////////////////////////////////////////////////////////////
// Error handling.

METHODDEF(void)
error_exit (j_common_ptr pcinfo)
// This procedure is called by the IJPEG library when an error
// occurs.
{
  /* Create the message string */
  char sz[256];
  (pcinfo->err->format_message) (pcinfo, sz);
  strcat (sz, "\n");

  CPicDecoder::raiseError (ERR_FORMAT_NOT_SUPPORTED, sz);
}

/////////////////////////////////////////////////////////////////////
// Class functions

CJPEGDecoder::CJPEGDecoder
    ()
    // Creates a decoder
{
  cinfo.err = jpeg_std_error (&jerr);
  jerr.error_exit = error_exit;  // Register custom error manager.

  jpeg_create_decompress (&cinfo);

  m_bFast = TRUE;
}


CJPEGDecoder::~CJPEGDecoder
    ()
{
  jpeg_destroy_decompress (&cinfo);
}

void CJPEGDecoder::SetFast
    ( BOOL bFast
    )
    // TRUE (the default) selects fast but sloppy decoding.
{
  m_bFast = bFast;
}

/* I used this function to test the IJPEG rouines without a custom
   data source. If something goes wrong, it may still be useful to
   isolate bugs.

CBmp * CJPEGDecoder::MakeBmpFromFile
    ( char * pszFName
    )
{
  char sz[256];
  sprintf (sz, "Decoding file %s.\n", pszFName);
  trace (1, sz);

  FILE * pFile;

  if ((pFile = fopen (pszFName, "rb")) == NULL)
  {
    TRACE ("Couldn't open file.\n");
  }

  try
  {
    jpeg_stdio_src (&cinfo, pFile);

    jpeg_read_header (&cinfo, TRUE);

    int w = cinfo.image_width;
    int h = cinfo.image_height;

    pBmp->Create (w, h, 32, FALSE);

    jpeg_start_decompress (&cinfo);

    if (cinfo.out_color_space == JCS_GRAYSCALE)
      decodeGray (w, h);
     else
      decodeRGB (w, h);
    jpeg_finish_decompress (&cinfo);
  }
  catch (CTextException)
  {
    jpeg_abort_decompress(&cinfo);
    fclose (pFile);
    throw;
  }
  fclose (pFile);
  return pBmp;
}
*/

void CJPEGDecoder::DoDecode
    ( CBmp * pBmp,
      RGBAPIXEL ** ppPal,
      int * pDestBPP,
      CDataSource * pDataSrc
    )
{
  try
  {
	// Jo Hagelberg 15.4.99: added progress notification callback
    JMETHOD( void, notify, (j_common_ptr));
    notify = JNotification;

    // Initialize custom data source.
    // Jo Hagelberg 15.4.99: added pDataSrc and notify
    jpeg_mem_src (&cinfo, pDataSrc->ReadEverything(),
                  pDataSrc->GetFileSize(), (void*)pDataSrc, notify);

    jpeg_read_header (&cinfo, TRUE);

    if (m_bFast)
    {
      cinfo.do_fancy_upsampling = FALSE;
    }

    // Choose floating point DCT method.
    cinfo.dct_method = JDCT_FLOAT;

    int w = cinfo.image_width;
    int h = cinfo.image_height;

    jpeg_start_decompress (&cinfo);

    if (cinfo.out_color_space == JCS_GRAYSCALE)
      decodeGray (pBmp, w, h, pDestBPP);
    else
      decodeRGB (pBmp, w, h, pDestBPP);
    jpeg_finish_decompress (&cinfo);
  }
  catch (CTextException)
  {
    jpeg_abort_decompress(&cinfo);
    throw;
  }


}

void CJPEGDecoder::decodeRGB
    ( CBmp * pBmp,
      int w,
      int h,
      int * pDestBPP
    )
    // Assumes IJPEG decoder is already set up.
{
  BYTE * pDst;
  int CurLine = 0;
  JSAMPARRAY ppDst = &pDst;

  CalcDestBPP (32, pDestBPP);
  pBmp->Create (w, h, 32, FALSE);

  BYTE ** pLineArray = pBmp->GetLineArray();

  while (CurLine < h)
  {
    pDst = pLineArray[CurLine];
    jpeg_read_scanlines (&cinfo, ppDst, 1);
    CurLine++;
  }
}

void CJPEGDecoder::decodeGray
    ( CBmp * pBmp,
      int w,
      int h,
      int * pDestBPP
    )
    // Assumes IJPEG decoder is already set up.
{
  BYTE * pDst;
  int CurLine = 0;
  BYTE * pBuf = new BYTE [w];
  JSAMPARRAY ppBuf = &pBuf;

  CalcDestBPP (8, pDestBPP);
  pBmp->Create (w, h, *pDestBPP, FALSE);

  BYTE ** pLineArray = pBmp->GetLineArray();

  while (CurLine < h)
  {
    if (*pDestBPP == 32)
    {
      jpeg_read_scanlines (&cinfo, ppBuf, 1);

      pDst = pLineArray[CurLine];

      for (int i=0; i<w; i++)
      {
        pDst[i*4] = pBuf[i];
        pDst[i*4+1] = pBuf[i];
        pDst[i*4+2] = pBuf[i];
        pDst[i*4+3] = 0xFF;
      }
    }
    else
    {
      ppBuf = &pDst;
      *ppBuf = pLineArray[CurLine];
      jpeg_read_scanlines (&cinfo, ppBuf, 1);

    }
    CurLine++;
  }
  delete pBuf;
}


/*
 * Jo Hagelberg 15.4.99
 * progress notification callback
 * since this is a static function we need pDataSrc from cinfo->client_data
 * progress is calculated (0...1) and PaintLib's Notification called
 */

void CJPEGDecoder::JNotification (j_common_ptr cinfo)
{
  double       part;
  CDataSource *pDataSrc;

  // set lock for threadsafety
#ifdef _WINDOWS
  // *this is not tested yet*
  static CCriticalSection JNotifyCS;
  CSingleLock   notifylock( &JNotifyCS, TRUE);
#endif

  /* calculated according to jpeg lib manual
   * note: this may not be precice when using buffered image mode
   * todo: think hard of alternatives 4 this case ... :-)
   */
  part = ( (double)cinfo->progress->completed_passes +
 	         ((double)cinfo->progress->pass_counter/cinfo->progress->pass_limit) ) /
	       (double)cinfo->progress->total_passes;

  // call Notification in CDataSource
  if( pDataSrc = (CDataSource*) cinfo->client_data)
  {
    pDataSrc->OProgressNotification( part);
  }

  // lock is unlocked at destruction :-)
}

/*
/--------------------------------------------------------------------
|
|      $Log: jpegdec.cpp,v $
|      Revision 1.5  1999/10/19 21:25:16  Ulrich von Zadow
|      no message
|
|      Revision 1.4  1999/10/03 18:50:51  Ulrich von Zadow
|      Added automatic logging of changes.
|
|
--------------------------------------------------------------------
*/
